#include "StdAfx.h"
#include "CadMath.h"

#include <math.h>

CadMath::CadMath(void)
{
}

CadMath::~CadMath(void)
{
}


RVector::RVector(double vx, double vy, double vz) :
    x(vx), y(vy), z(vz) {
}

/**
 * \return Magnitude (length) of the vector.
 */
double RVector::getMagnitude() const {
    // Note that the z coordinate is also needed for 2d
    //   (due to definition of crossP())
    return sqrt(pow(x, 2) + pow(y, 2) + pow(z, 2));
}

/**
 * binary + operator.
 */
RVector RVector::operator +(const RVector& v) const {
    return RVector(x + v.x, y + v.y, z + v.z);
}

/**
 * binary - operator.
 */
RVector RVector::operator -(const RVector& v) const {
    return RVector(x - v.x, y - v.y, z - v.z);
}

/**
 * binary * operator.
 */
RVector RVector::operator *(double s) const {
    return RVector(x * s, y * s, z * s);
}

/**
 * binary / operator.
 */
RVector RVector::operator /(double s) const {
    return RVector(x / s, y / s, z / s);
}

/**
 * unary - operator.
 */
RVector RVector::operator -() const {
    return getNegated();
}

/**
 * \return New vector with negated components.
 */
RVector RVector::getNegated() const {
    return RVector(-x, -y, -z);
}

/**
 * Scalarproduct (dot product).
 */
double RVector::getDotProduct(const RVector& v1, const RVector& v2) {
    return v1.x * v2.x + v1.y * v2.y + v1.z * v2.z;
}

RVector RVector::getCrossProduct(const RVector& v1, const RVector& v2) 
{
	return RVector(v1.y * v2.z - v1.z * v2.y, v1.z * v2.x - v1.x * v2.z, v1.x * v2.y - v1.y * v2.x);
}

double RVector::Dot(RVector& v)
{
	return getDotProduct(*this, v);
}

RVector RVector::Cross(RVector& v)
{
	return getCrossProduct(*this, v);
}

/**
 * += operator. The result is only valid if both vectors are valid.
 */
void RVector::operator +=(const RVector& v) {
    x += v.x;
    y += v.y;
    z += v.z;
}

/**
 * -= operator. The result is only valid if both vectors are valid.
 */
void RVector::operator -=(const RVector& v) {
    x -= v.x;
    y -= v.y;
    z -= v.z;
}

/**
 * *= operator
 */
void RVector::operator *=(double s) {
    x *= s;
    y *= s;
    z *= s;
}

/**
 * /= operator
 */
void RVector::operator /=(double s) {
    x /= s;
    y /= s;
    z /= s;
}

/**
 * == operator
 */
bool RVector::operator ==(const RVector& v) const {
    return x == v.x && y == v.y && z == v.z;
}

/**
 * \return Magnitude (length) of the vector projected to the x/y plane (2d).
 */
double RVector::getMagnitude2d() const {
    return sqrt(pow(x, 2) + pow(y, 2));
}

void RVector::setPolar(double radius, double angle) {
    x = radius * cos(angle);
    y = radius * sin(angle);
    z = 0.0;
}

void RVector::setAngle(double a) {
    double m = getMagnitude();
    setPolar(m, a);
}
#ifndef M_PI  //add by linfs 
#define M_PI       3.14159265358979323846
#endif
/**
 * \return The angle from zero to this vector (in rad).
 */
double RVector::getAngle() const {
    double ret = 0.0;
    double m = getMagnitude2d();

    if (m > 1.0e-6) {
        double dp = getDotProduct(*this, RVector(1.0, 0.0));
        if (dp / m >= 1.0) {
            ret = 0.0;
        } else if (dp / m < -1.0) {
            ret = M_PI;
        } else {
            ret = acos(dp / m);
        }
        if (y < 0.0) {
            ret = 2*M_PI - ret;
        }
    }
    return ret;
}

/**
 * \return The angle from this and the given coordinate (in rad).
 */
double RVector::getAngleTo(const RVector& v) const {
       return (v - (*this)).getAngle();
}

double RVector::getAngle3dTo(const RVector& v) const{
	// add by houzz at 13:51 2018/10/18
	double ret = 0.0;
	double m = getMagnitude()*v.getMagnitude();

	if (m > 1.0e-6) {
		double dp = getDotProduct(*this, v);
		if (dp / m >= 1.0) {
			ret = 0.0;
		}
		else if (dp / m < -1.0) {
			ret = M_PI;
		}
		else {
			ret = acos(dp / m);
		}
		if (dp < 0.0) {
			ret = 2 * M_PI - ret;
		}
	}
	return ret;
}

/**
 * \return The distance between this and the given coordinate.
 */
double RVector::getDistanceTo(const RVector& v) const {
     return (*this - v).getMagnitude();
}

/**
* Rotates this vector around 0/0 by the given angle.
*/
RVector RVector::rotate(double rotation) {

	double r = getMagnitude2d();
	double a = getAngle() + rotation;

	x = cos(a) * r;
	y = sin(a) * r;

	return *this;
}

/**
* Rotates this vector around the given center by the given angle.
*/
RVector RVector::rotate(double rotation, const RVector& center) {
	*this = center + (*this - center).rotate(rotation);
	return *this;
}

//㷨Դhttp://blog.csdn.net/qiuchangyong/article/details/5859628
RVector RVector::rotate3d(double rotation, RVector rvDir)
{
#if 0
	RVector dir = rvDir;
	double dl = dir.getMagnitude();
	dir.x = rvDir.x / dl;
	dir.y = rvDir.y / dl;
	dir.z = rvDir.z / dl;

	double mm[4][4] = { 0 };
	double fCos = cos(rotation);

	double fSin = sin(rotation);



	mm[0][0] = ( dir.x * dir.x ) * ( 1.0f - fCos ) + fCos;

	mm[0][1] = (dir.x * dir.y) * (1.0f - fCos) - (dir.z * fSin);

	mm[0][2] = (dir.x * dir.z) * (1.0f - fCos) + (dir.y * fSin);



	mm[1][0] = (dir.y * dir.x) * (1.0f - fCos) + (dir.z * fSin);

	mm[1][1] = (dir.y * dir.y) * (1.0f - fCos) + fCos;

	mm[1][2] = (dir.y * dir.z) * (1.0f - fCos) - (dir.x * fSin);



	mm[2][0] = (dir.z * dir.x) * (1.0f - fCos) - (dir.y * fSin);

	mm[2][1] = (dir.z * dir.y) * (1.0f - fCos) + (dir.x * fSin);

	mm[2][2] = (dir.z * dir.z) * (1.0f - fCos) + fCos;



	mm[0][3] = mm[1][3] = mm[2][3] = 0;

	mm[3][0] = mm[3][1] = mm[3][2] = 0;

	mm[3][3] = 1;


	double a = mm[0][0];
	double b = mm[0][1];
	double c = mm[0][2];
	double d = mm[0][3];
	double e = mm[1][0];
	double f = mm[1][1];
	double g = mm[1][2];
	double h = mm[1][3];
	double i = mm[2][0];
	double j = mm[2][1];
	double k = mm[2][2];
	double l = mm[2][3];
	double m = mm[3][0];
	double n = mm[3][1];
	double o = mm[3][2];
	double p = mm[3][3];

	double x1 = x, y1 = y, z1 = z;
	double x2 = a*x1 + b*y1 + c*z1 + 0;
	double y2 = e*x1 + f*y1 + g*z1 + 0;
	double z2 = i*x1 + j*y1 + k*z1 + 0;

	x = x2;
	y = y2;
	z = z2;

#elif 0
	double c = cos(rotation);
	double s = sin(rotation);


	double px = x, py = y, pz = z;
	double ax = rvDir.x, ay = rvDir.y, az = rvDir.z;


	double xx = px * c+( ay * pz- az * py)*s + ax *(ax * px + ay * py + az * pz)*(1 - c);
	double yy = py * c+( ax * pz- az * px)*s + ay *(ax * px + ay * py + az * pz)*(1 - c);
	double zz = pz * c+( ax * py- ay * px)*s + az* (ax * px + ay * py + az * pz)*(1 - c);

	x = xx;
	y = yy;
	z = zz;

#else

	double c = cos(rotation);
	double s = sin(rotation);

	double old_x = x;
	double old_y = y;
	double old_z = z;

	//ʸת
	double dl = rvDir.getMagnitude();
	x = rvDir.x / dl;
	y = rvDir.y / dl;
	z = rvDir.z / dl;

	double dx = (x*x*(1 - c) + c) * old_x + (x*y*(1 - c) - z*s) * old_y + (x*z*(1 - c) + y*s) * old_z;
	double dy = (y*x*(1 - c) + z*s) * old_x + (y*y*(1 - c) + c) * old_y + (y*z*(1 - c) - x*s) * old_z;
	double dz = (x*z*(1 - c) - y*s) * old_x + (y*z*(1 - c) + x*s) * old_y + (z*z*(1 - c) + c) * old_z;
	x = dx;
	y = dy;
	z = dz;
#endif

	return *this;
}

RVector RVector::stretch3d(RVector rvFrom, RVector rvTo)
{
	// add by houzz at 13:51 2018/10/18
	// ȷԻҪ֤

	// rvFrom쵽rvTo

	// 
	// һ(x1, y1, z1)ӦĬĬΪ(0, 0, 1)(x1, y1, z1)
	// (-0.0023609854512506, 0.002477970246664, 0.9999941426884237), ӦӦΪ

	if (fabs(rvFrom.x - rvTo.x) < 1e-6
		&& fabs(rvFrom.y - rvTo.y) < 1e-6
		&& fabs(rvFrom.z - rvTo.z) < 1e-6)
	{
		return *this;
	}

	double dv1 = sqrt(rvFrom.x*rvFrom.x + rvFrom.y*rvFrom.y + rvFrom.z*rvFrom.z);
	double dv2 = sqrt(rvTo.x*rvTo.x + rvTo.y*rvTo.y + rvTo.z*rvTo.z);
	if (dv1 < 1e-6 || dv2 < 1e-6)
	{
		return *this;
	}

	double dFac = dv2 / dv1;
	if (fabs(dFac - 1)> 1e-6)
	{
		*this *= dFac;
	}

	double l = rvFrom.x, m = rvFrom.y, n = rvFrom.z;
	double o = rvTo.x, p = rvTo.y, q = rvTo.z;
	RVector rf(m*q - n*p, n*o - l*q, l*p - m*o);

	double r = rvFrom.getAngle3dTo(rvTo);

#if defined DEBUG && 0
	double d = TO_DEG(r);
	RVector test = *this;
	test.rotate3d(2 * M_PI - r, rf);

	// RVector test1 = *this;
	// test1.rotate3d(M_PI, rf);
#endif

	this->rotate3d(r, rf);

	return *this;
}

//rvLineStartrvLineEndת
RVector RVector::rotate3d(double rotation, RVector rvLineStart, RVector rvLineEnd)
{
	*this -= rvLineStart;
	rotate3d(rotation, rvLineEnd - rvLineStart);
	*this += rvLineStart;
	return *this;
}

RVector RVector::GetLatLngAlt(double dRadius)
{
	double dd = getMagnitude();
	double dAlt = dd - dRadius;
	double lat = -asin(y / dd) * 180 / M_PI;
	double lng = -(atan2(z, x) * 180 / M_PI);
	RVector rvLl(lng, lat, dAlt);
	return rvLl;
}

#define RV_RAD(d) (d*M_PI / 180.0)
double GetDistanceInBall(double lng1, double lat1, double lng2, double lat2, double dRadius)
{
	double radLat1 = RV_RAD(lat1);
	double radLat2 = RV_RAD(lat2);
	double a = radLat1 - radLat2;
	double b = RV_RAD(lng1) - RV_RAD(lng2);
	double s = 2 * asin(sqrt(pow(sin(a / 2), 2) + cos(radLat1)*cos(radLat2)*pow(sin(b / 2), 2)));
	s = s * dRadius;
	return s;
}

//ԭƶrvToתǶ
double RVector::GetRotateDirAndAngle(RVector rvTo, RVector& rvDir)
{
	double dRadius = getMagnitude();
	// double ddd = rvTo.getMagnitude();
	RVector rvLlFrom = GetLatLngAlt(dRadius);
	RVector rvLlTo = rvTo.GetLatLngAlt(dRadius);
	double dd = GetDistanceInBall(rvLlFrom.x, rvLlFrom.y, rvLlTo.x, rvLlTo.y, dRadius);
	double angle = dd / dRadius;
	rvTo.scale(1 / cos(angle), RVector(0, 0, 0));
	rvTo.rotate3d(M_PI / 2, *this);
	rvDir = *this - rvTo;
	return angle;
}

//rv0ƶrvToתǶ
double RVector::GetRotateDirAndAngle(RVector rv0, RVector rvTo, RVector& rvDir)
{
	RVector rvMe = *this - rv0;
	rvTo -= rv0;
	double angle = rvMe.GetRotateDirAndAngle(rvTo, rvDir);
	rvDir += rv0;
	double dd0 = rvDir.getMagnitude();
	double dd = rvDir.getDistanceTo(rv0);
	double d0 = rv0.getMagnitude();
	if (dd0 < dd || dd0 < d0)
	{
		angle = -angle;
		rvDir.scale(-1, RVector());
	}
	return angle;
}

/**
* Linear interpolation between this and \c v by fraction 't'.
*/
RVector RVector::getLerp(const RVector& v, double t) const {
	return RVector(x + (v.x - x) * t, y + (v.y - y) * t, z + (v.z - z) * t);
}

double RVector::getDistanceToLine(const RVector& v1, const RVector& v2) const
{
	if (v1.x == v2.x)
		return fabs(x - v1.x);
	if (v1.y == v2.y)
		return fabs(y - v1.y);

	double b1 = (v2.y - v1.y) / (v2.x - v1.x);
	double b2 = -1 / b1;
	double a1 = v1.y - b1*v1.x;
	double a2 = y - b2*x;

	double dx = (a1 - a2) / (b2 - b1);
	double dy = a2 + b2*dx;
	RVector vInter(dx, dy);
	return getDistanceTo(vInter);
}

//㵽ߵľ룬v1dAngleʾֱ
double RVector::getDistanceToLine1(const RVector& v1, double dAngle)
{
	if (fabs(dAngle) <DOUBLE_EQ_DIFF || fabs(fabs(dAngle) - M_PI) < DOUBLE_EQ_DIFF || fabs(fabs(dAngle) - M_PI_360) < DOUBLE_EQ_DIFF)
		return fabs(y - v1.y);
	if (fabs(fabs(dAngle) - M_PI_90) < DOUBLE_EQ_DIFF || fabs(fabs(dAngle) - M_PI_270) < DOUBLE_EQ_DIFF)
		return fabs(x - v1.x);

	double b1 = tan(dAngle);
	double b2 = -1 / b1;
	double a1 = v1.y - b1*v1.x;
	double a2 = y - b2*x;

	double dx = (a1 - a2) / (b2 - b1);
	double dy = a2 + b2*dx;
	RVector vInter(dx, dy);
	return getDistanceTo(vInter);
}

//ȡĵ꣬v1v2Ϊһϵ㣬ѵĵΪӦϵ
RVector RVector::getMiddleVt(const RVector& v1, const RVector& v2)
{
	if (v1.x == v2.x)
		return RVector((x+v1.x)/2, (v1.y+v2.y)/2, 0);
	if (v1.y == v2.y)
		return RVector((v1.x + v2.x) / 2, (y + v2.y) / 2, 0);

	double midx = (v1.x + v2.x) / 2;
	double midy = (v1.y + v2.y) / 2;

	double b1 = (v2.y - v1.y) / (v2.x - v1.x);
	double b2 = -1 / b1;
	double a1 = y - b1*x;
	double a2 = midy - b2*midx;

	double dx = (a1 - a2) / (b2 - b1);
	double dy = a2 + b2*dx;
	return RVector((dx + midx) / 2, (dy + midy) / 2);
}

RVector RVector::scaleF(double factor, const RVector& center)
{
	x = (x - center.x)*factor + center.x;
	y = (y - center.y)*factor + center.y;
	z = (z - center.z)*factor + center.z;
	return *this;
}

RVector RVector::scale(double factor, const RVector& center)
{
	//change by houyb,ԭΪеݹãٶȼ(Լ10)
	return scaleF(factor, center);

	if (center == RVector()) {
		x *= factor;
		y *= factor;
		z *= factor;
		return *this;
	}

	*this = center + (*this - center).scale(factor, RVector(0,0,0));
	return *this;
}

double RVector::GetScale(RVector& rvCenter, RVector& rvOther)
{
	bool bNearCenter = false;
	double dOtherToCenter = rvOther.getDistanceTo(rvCenter);
	double dOtherToMe = rvOther.getDistanceTo(*this);
	double dMeToCenter = getDistanceTo(rvCenter);
	if (dOtherToCenter < dOtherToMe)
		bNearCenter = true;
	if (bNearCenter)
		return -dOtherToCenter / dMeToCenter;
	else
		return dOtherToCenter / dMeToCenter;
}

//ÿ150000(debug10000)
void RVector::MakeEarthXyz(double lat, double lng, double dRadius)
{
	double a = (-lng) / 180 * M_PI;
	x = cos(a) * dRadius;
	y = sin(a) * dRadius;

	double da = lat / 180 * M_PI;
	double ds = cos(da);
	scale(ds, RVector());
	z = (dRadius)*sin(da);

	double d = z;
	z = y;
	y = -d;
}

//潻㣬Ϊ㣬rvToΪ߷dBallRadiusΪ뾶rvNearΪ㣬ֵǷн
bool RVector::GetCrossToBall(RVector rvTo, double dBallRadius, RVector& rvNear)
{
	//Ҷ Ϊa, b, c.ΪA, B, C.
	//Ҷa / sinA = b / sinB = c / sinC

	double dMinDouble = 0.00001;
	rvNear = RVector();
	double dMe = this->getMagnitude();

	if (dMe <= dBallRadius + dMinDouble)
	{
		return false;
	}

	rvTo.scale(dMe / this->getDistanceTo(rvTo), *this);

	RVector rvDir;
	RVector rv0;
	double dAngleA = fabs(rv0.GetRotateDirAndAngle(*this, rvTo, rvDir));
	if (dAngleA < 0.0000000001)
	{
		rvNear = *this;
		rvNear.scale(dBallRadius / dMe, RVector());
		return true;
	}

	double dAngleT = asin(dBallRadius / dMe);
	if (dAngleA >= dAngleT)
		return false;

	double dAngleB1 = asin(dMe / (dBallRadius / sin(dAngleA)));
	double dAngleB2 = M_PI - dAngleB1;
	double dAngleC1= M_PI - dAngleA - dAngleB1;
	double dAngleC2 = M_PI - dAngleA - dAngleB2;
	double dc1 = (dBallRadius / sin(dAngleA)) * sin(dAngleC1);
	double dc2 = (dBallRadius / sin(dAngleA)) * sin(dAngleC2);

	if (dAngleC1 < 0)
		dAngleC1 = dAngleC1;

	RVector rv1 = rvTo;
	double ds = dc1 / dMe;
	rv1.scale(ds, *this);
	RVector rv2 = rvTo;
	ds = dc2 / dMe;
	rv2.scale(ds, *this);

	rvNear = rv1;
	if (rv1.getDistanceTo(*this) > rv2.getDistanceTo(*this))
		rvNear = rv2;

	return true;
}

RArc::RArc() :
    center(RVector(0,0,0)),
    radius(0.0),
    startAngle(0.0),
    endAngle(0.0),
    reversed(false) {
}

RArc::RArc(const RVector& center, double radius, double startAngle,
        double endAngle, bool reversed) :
    center(center),
    radius(radius),
    startAngle(startAngle),
    endAngle(endAngle),
    reversed(reversed) {
}

RArc RArc::createFrom2PBulge(const RVector& startPoint,
        const RVector& endPoint, double bulge) {

	RArc arc;

	arc.reversed = (bulge < 0.0);
	double alpha = atan(bulge) * 4.0;

	RVector middle = (startPoint + endPoint) / 2.0;
	double dist = startPoint.getDistanceTo(endPoint) / 2.0;

	// alpha can't be 0.0 at this point
	arc.radius = fabs(dist / sin(alpha / 2.0));

	double wu = fabs(pow(arc.radius, 2.0) - pow(dist, 2.0));
	double h = sqrt(wu);
	double angle = startPoint.getAngleTo(endPoint);

	if (bulge > 0.0) {
		angle += M_PI / 2.0;
	} else {
		angle -= M_PI / 2.0;
	}

	if (fabs(alpha) > M_PI) {
		h *= -1.0;
	}

	arc.center.setPolar(h, angle);
	arc.center += middle;
	arc.startAngle = arc.center.getAngleTo(startPoint);
	arc.endAngle = arc.center.getAngleTo(endPoint);

	return arc;
}

/**
* Creates an arc from 3 points.
*/
RArc RArc::createFrom3Points(const RVector& rvStart, const RVector& rv, const RVector& rvEnd) 
{
	RVector rvL1s = rvStart;
	rvL1s.scale(0.5, rv);
	RVector rvL1e = rvStart;
	rvL1e.rotate(M_PI / 2, rvL1s);

	RVector rvL2s = rvEnd;
	rvL2s.scale(0.5, rv);
	RVector rvL2e = rvEnd;
	rvL2e.rotate(M_PI / 2, rvL2s);

	RVector rvInt;
	rvInt.x = rvInt.y = PT_INVALID_XY;
	bool b = TwoLineInteractionRv(rvL1s, rvL1e, rvL2s, rvL2e, rvInt);
	(void)b;
	if (rvInt.x == PT_INVALID_XY && rvInt.y == PT_INVALID_XY)
	{
		return RArc();
	}

	double dRadius = rvInt.getDistanceTo(rvStart);

	double dA = rvStart.getAngleTo(rvInt);
	double dB = rv.getAngleTo(rvInt);
	double dC = rvEnd.getAngleTo(rvInt);
	bool bNeedChange = false;
	if (dA < dC)
	{
		if (dB<dA || dB>dC)
			bNeedChange = true;
	}
	else
	{
		if (dB<dA && dB>dC)
			bNeedChange = true;
	}

	if (bNeedChange)
	{
		dB = dA;
		dA = dC;
		dC = dB;
	}

	return RArc(rvInt, dRadius, dA, dC);
}

RVector RArc::getPoint(double dAngle) const
{
	return RVector(center.x + cos(dAngle) * radius, center.y + sin(
		dAngle) * radius);
}

RVector RArc::getStartPoint() const
{
    return RVector(center.x + cos(startAngle) * radius, center.y + sin(
            startAngle) * radius);
}

RVector RArc::getEndPoint() const
{
    return RVector(center.x + cos(endAngle) * radius, center.y + sin(
            endAngle) * radius);
}
    
double RArc::getSweep() const
{
    double ret = 0.0;

    if (endAngle <= startAngle) {
        ret = endAngle + 2 * M_PI - startAngle;
    } else {
        ret = endAngle - startAngle;
    }

    return ret;
}

double RArc::getBulge() const
{
    double bulge = tan(fabs(getSweep()) / 4.0);
    return bulge;
}

double GetEllipseLength(double a, double b)
{
	// ƹʽ
	// a, b: ̰
	if (a == 0 || b == 0)
		return 0;
	if (a < 0)
		a = -a;
	if (b < 0)
		b = -b;
	return M_PI*(1.5*(a + b) - sqrt(a*b));
}

double GetEllipseSimLength(double a, double b)
{
	// ֿ, ֱģ
	if (a == 0 || b == 0)
		return 0;
	if (a < 0)
		a = -a;
	if (b < 0)
		b = -b;
	// a, b: ̰
	int iSplit = 360;

	double prex = a, prey = 0;
	double dAxisScale = b / a;
	double dDist = 0;

	RVector rvCenter(0, 0);
	RVector rvEnd(a, 0);

	for (int i = 1; i <= iSplit; i++)
	{
		double dAngle = iSplit*i / iSplit;
		dAngle = 180 - dAngle;
		RVector rvXy = GetXyInTiltEllipse(rvCenter, rvEnd, dAxisScale, dAngle);
		dDist += GetLineDist(rvXy.x, rvXy.y, prex, prey);
		prex = rvXy.x;
		prey = rvXy.y;
	}

	return dDist;
}

double GetEllipseArcLength(double a, double b, double dStartangle, double dEndangle, double dLenCalc, double dLenSim)
{
	if (a == 0 || b == 0)
		return 0;
	// ѧǶ
	if (a < 0)
		a = -a;
	if (b < 0)
		b = -b;

	if (dLenCalc < 0)
		dLenCalc  = GetEllipseLength(a, b);
	if (dLenSim < 0)
		dLenSim = GetEllipseSimLength(a, b);


	// ֿ, ֱģ
	if (a == 0 || b == 0)
		return 0;
	// a, b: ̰
	// int iSplit = 360;
	// int iSplit = 72;
	int iSplit = (int)ceil(dEndangle - dStartangle);
	if (iSplit > 360)
		iSplit = 360;

	dEndangle = StandrandAngle(dEndangle, dStartangle);
	RVector rvCenter(0, 0);
	RVector rvEnd(a, 0);
	double dAxisScale = b / a;

	RVector rvPrev = GetXyInTiltEllipse(rvCenter, rvEnd, dAxisScale, 180 - dStartangle);
	double dDist = 0;

	for (int i = 1; i <= iSplit; i++)
	{
		double dAngle = dStartangle + (dEndangle - dStartangle)*i / iSplit;
		dAngle = 180 - dAngle;

		RVector rvXy = GetXyInTiltEllipse(rvCenter, rvEnd, dAxisScale, dAngle);
		dDist += GetLineDist(rvXy.x, rvXy.y, rvPrev.x, rvPrev.y);
		rvPrev = rvXy;
	}

	if (dLenCalc > dLenSim && dLenSim > 0)
		dDist = dDist*dLenCalc / dLenSim;

	return dDist;
}

double GetEllipseArcAngle(double a, double b, double dStartangle, double dLength, double dLenCalc, double dLenSim)
{
	if (dLenCalc < 0)
		dLenCalc = GetEllipseLength(a, b);
	if (dLenSim < 0)
		dLenSim = GetEllipseSimLength(a, b);

	double dMinAngle = dStartangle;
	double dMaxAngle = dStartangle + 360;
	double dCurAngle = (dMinAngle + dMaxAngle) / 2;

	double dCurLen = GetEllipseArcLength(a, b, dStartangle, dCurAngle, dLenCalc, dLenSim);

	// 100
	int iMaxCnt = 100;
	while (iMaxCnt > 0 && fabs(dLength - dCurLen)>1e-6)
	{
		if (dLength > dCurLen)
		{
			dMinAngle = dCurAngle;
		}
		else if (dLength < dCurLen)
		{
			dMaxAngle = dCurAngle;
		}
		else
		{
			return dCurAngle;
		}

		dCurAngle = (dMinAngle + dMaxAngle) / 2;
		if (fabs(dMaxAngle-dMinAngle) < 1e-6)
			return dCurAngle;
	
		dCurLen = GetEllipseArcLength(a, b, dStartangle, dCurAngle, dLenCalc, dLenSim);
		iMaxCnt--;
	}

	return dCurAngle;
}

CRect64 Rect32To64(RECT &rect)
{
	return CRect64(rect.left, rect.top, rect.right, rect.bottom);
}

CRect Rect64To32(RECTH64 &rect)
{
	return CRect((int)rect.left, (int)rect.top, (int)rect.right, (int)rect.bottom);
}


RVector GetXyInTiltEllipse(RVector rvCenter, RVector rvEnd, double dAxisScale, double dAngle)
{
	RVector rvXy;
	GetXyInTiltEllipse(rvCenter, rvEnd, dAxisScale, dAngle, rvXy.x, rvXy.y);
	return rvXy;
}

void GetXyInTiltEllipse(RVector rvCenter, RVector rvEnd, double dAxisScale, double dAngle, double& x, double &y)
{
	x = y = 0;
	if (dAxisScale < 0)
		return;
	if (dAxisScale == 0)
		dAxisScale = 1;
	dAngle = dAngle*M_PI / 180;
	double aAxis = rvEnd.getDistanceTo(rvCenter);
	double bAxis = aAxis*dAxisScale;
	if (aAxis <= 0 || bAxis <= 0)
	{
		x = rvCenter.x;
		y = rvCenter.y;
		return;
	}

	double ang = StandrandRadian(dAngle);
	if (fabs(ang - M_PI / 2) < 1e-6 || fabs(ang - M_PI / 2 * 3) < 1e-6)
	{
		x = 0;
		if (fabs(ang - M_PI / 2) < 1e-6)
			y = bAxis;
		else
			y = -bAxis;
	}
	else
	{
		double k = tan(ang);
		x = sqrt(1 / (1 / (aAxis*aAxis) + (k*k) / (bAxis*bAxis)));
		if (ang > M_PI / 2 && ang < M_PI / 2 * 3)
			x = -x;
		y = k*x;
	}

	RVector rv(x, y);
	rv.rotate(rvEnd.getAngleTo(rvCenter));
	rv += rvCenter;
	x = rv.x;
	y = rv.y;
}

void GetXyInEllipse64(CRect64 reArea, double dAngle, double& x, double &y)
{
	// dAngle, ѧʵĽǶ
	dAngle = dAngle*M_PI / 180;
	double a = (double)reArea.Width();
	double b = (double)reArea.Height();
	a = a / 2;
	b = b / 2;
	if (reArea.Width() == 0 || reArea.Height() == 0)
	{
		x = reArea.left + a;
		y = reArea.top + b;
		return;
	}

	double ang = M_PI - dAngle;
	while (ang < -M_PI)
		ang += 2 * M_PI;
	while (ang > M_PI)
		ang -= 2 * M_PI;

	if (fabs((fabs(ang) - M_PI / 2)) < 1e-6)
	{
		x = 0;
		if (ang > 0)
			y = b;
		else
			y = -b;
	}
	else
	{
		double k = tan(ang);
		x = sqrt(1 / (1 / (a*a) + (k*k) / (b*b)));
		if (ang>M_PI / 2 || ang<-M_PI / 2)
			x = -x;
		y = k*x;
	}

	// y᷽෴
	y = -y;

	// double cx = (double)((reArea.left + reArea.right)/2);
	// double cy = (double)((reArea.top + reArea.bottom)/2);
	// 16:32 2018/8/3
	double cx = (reArea.left + reArea.right)/2.0;
	double cy = (reArea.top + reArea.bottom)/2.0;
	x += cx;
	y += cy;
}

void GetXyInEllipse(CRect reArea, double dAngle, double& x, double &y)
{
	GetXyInEllipse64(Rect32To64(reArea), dAngle, x, y);
}

double GetEllipseParamAngle64(CRect64 reArea, double dAngle)
{
	double x = 0, y = 0;
	double cx = (reArea.left + reArea.right) / 2.0;
	// double cy = (reArea.top + reArea.bottom) / 2.0;

	double a2 = (reArea.right - reArea.left) / 2.0;
	if (fabs(a2) < 1e-6)
		return 0;

	GetXyInEllipse64(reArea, dAngle, x, y);
	double ddValue = (x - cx) / a2;
	if (ddValue < -1)
		ddValue = -1;
	if (ddValue > 1)
		ddValue = 1;
	double ddRad = M_PI - acos(ddValue);
	dAngle = StandrandAngle(dAngle);
	if (dAngle > 180)
		ddRad = 2*M_PI - ddRad;
	double ddAngle = TO_DEG(ddRad);

	return ddAngle;
}

void GetXyInEllipseParam64(CRect64 reArea, double dAngle, double& x, double &y)
{
	// dAngle, ʹõĽǶ, ƽ
	dAngle = dAngle*M_PI / 180;
	dAngle = M_PI - dAngle;

	double cx = (reArea.left + reArea.right) / 2.0;
	double cy = (reArea.top + reArea.bottom) / 2.0;

	double a = (double)reArea.Width();
	double b = (double)reArea.Height();
	a = a / 2;
	b = b / 2;

	x = a*cos(dAngle);
	y = b*sin(dAngle);

	// double k = atan2(y, x);

	y = -y;

	x += cx;
	y += cy;
}

bool GetRatioPoit2(double x1, double y1, double x2, double y2, double x3, double &y3)
{
	// ȡߵıֵ
	if (fabs(x1 - x2) < 1e-6)
	{
		y3 = y1;
		return true;
	}

	y3 = (y2 - y1)*(x3 - x1) / (x2 - x1) + y1;

	return true;
}

bool GetRatioPoit3(double x1, double y1, double z1, double x2, double y2, double z2, double x3, double &y3, double &z3)
{
	// ȡƽֵ
	double yy3 = 0, zz3 = 0;
	if (GetRatioPoit2(x1, y1, x2, y2, x3, yy3) == false)
		return false;

	if (GetRatioPoit2(y1, z1, y2, z2, yy3, zz3) == false)
		return false;

	y3 = yy3;
	z3 = zz3;
	return true;
}

#if 0
// ʹStandrandRadian(fRad, -M_PI)
double Get1PiRangeRad(double fRad)
{
	while (fRad > M_PI)
	{
		fRad -= 2 * M_PI;
	}
	while (fRad <= -M_PI)
	{
		fRad += 2 * M_PI;
	}

	return fRad;
}

// ʹStandrandAngle(fDegree, -180)
double Get1PiRangeDegree(double fDegree)
{
	while (fDegree > 180)
	{
		fDegree -= 360;
	}
	while (fDegree <= -180)
	{
		fDegree += 360;
	}

	return fDegree;
}

// ʹStandrandAngle(fDegree)
double Get2PiRangeDegree(double fDegree)
{
	while (fDegree >= 360)
	{
		fDegree -= 360;
	}
	while (fDegree < 0)
	{
		fDegree += 360;
	}

	return fDegree;
}
#endif


//	: (ԭ)ĳһϵת֮
//	˵:
//		dx1, dy1: ϵ1ĳ
//		dScale: ϵ2ű, 1.0: , <1: С, >1:Ŵ, עⷴ, ϵ2СӦԭӦǱ
//		dRotateDeg: ϵ2תǵĽǶֵ, ʱ뷽Ϊ, עⷴ, ϵ2ʱתx, ӦԭӦ˳ʱתx
//	ֵ:
//		bool
bool CsysTransformRestoration2f(double dx1, double dy1, double dScale, double dRotateDeg, double &dx2, double &dy2)
{
	dx2 = dy2 = 0;
	if (fabs(dScale) < 1e-6)
		return false;

	dx2 = dx1;
	dy2 = dy1;
	if (fabs(dScale - 1.0) >= 1e-6)
	{
		// С, ƶȨ(ƶͬľı仯), ó
		dx2 /= dScale;
		dy2 /= dScale;
	}

	dRotateDeg = StandrandAngle(dRotateDeg);
	// dRotateDeg = Get2PiRangeDegree(dRotateDeg);
	if (fabs(dRotateDeg) >= 1e-6)
	{
		CircleRotationDegFormula(dx2, dy2, -dRotateDeg);
	}



	return true;
}

//	: Բתʽ
void CircleRotationDegFormula(double &dx, double &dy, double dRotateDeg)
{
	double dRotateRad = TO_RAD(dRotateDeg);
	CircleRotationRadFormula(dx, dy, dRotateRad);
}

//	: Բתʽ
//	˵:
//		dx, dy[in, out]: Բϵ
//		adRotateRad: תǵĻֵ, ʱ뷽Ϊ
//	ֵ:
//		void
//	ע:
//		ʽԭ(0, 0)ת
void CircleRotationRadFormula(double &dx, double &dy, double dRotateRad)
{
	double dx1 = cos(dRotateRad)*dx - sin(dRotateRad)*dy;
	double dy1 = cos(dRotateRad)*dy + sin(dRotateRad)*dx;

	dx = dx1;
	dy = dy1;
}

void GetSquareMeanValue(double bl, double br, double tl, double tr, double fRatioX, double fRatioY, double &fMean)
{
	// ѧ, , ӦֵΪ
	// (0, 0) --> bl
	// (1, 0) --> br
	// (0, 1) --> tl
	// (1, 1) --> tr

	double bb = 0, tt = 0;
	GetRatioPoit2(0, bl, 1, br, fRatioX, bb);
	GetRatioPoit2(0, tl, 1, tr, fRatioX, tt);
	GetRatioPoit2(0, bb, 1, tt, fRatioY, fMean);
}


//ǶתΪ0-360֮
double StandrandAngle(double dAngle, double dBase, bool bUpBase)
{
	// bUpBaseʾֵdBase
	double dMin = dBase;
	double dMax = dBase + 360;

	while (dAngle < dMin || (bUpBase && dAngle == dMin))
		dAngle += 360;
	while ((dAngle - dMax)>1e-6)
		dAngle -= 360;
	return dAngle;
}

//ǶתΪ0-2PI֮
double StandrandRadian(double dRad, double dBase, bool bUpBase)
{
	// bUpBaseʾֵdBase
	double dMin = dBase;
	double dMax = dBase + 2 * M_PI;

	while (dRad < dMin || (bUpBase && dRad == dMin))
		dRad += 2 * M_PI;
	while ((dRad - dMax)>1e-6)
		dRad -= 2 * M_PI;
	return dRad;
}

double CadAngleTrans(double dAngle)
{
	dAngle=180-dAngle/M_PI*180;
	if(dAngle<0)
		dAngle+=360;
	return dAngle;
}

double GetLineAngle(double x1,double y1,double x2,double y2)
{
	RVector v1(x1,-y1);
	RVector v2(x2,-y2);
	return CadAngleTrans(v1.getAngleTo(v2));
}

double GetLineDist(double x1,double y1,double x2,double y2)
{
	double aa=sqrt((x1-x2)*(x1-x2)+(y1-y2)*(y1-y2));
	return aa;
}

double GetLineDistI64(INT64 x1, INT64 y1, INT64 x2, INT64 y2)
{
	INT64 xd = llabs(x1 - x2);
	INT64 yd = llabs(y1 - y2);
	INT64 xm = xd*xd;
	INT64 ym = yd*yd;

	if (xm < xd || ym < yd)
	{
		//
		return 10000000000000000l;
	}

	double aa = sqrt((double)xm + ym);
	return aa;
}

//㵽߶ε̾(pt3Ϊ)
double GetPointToLineDist(CPoint pt1, CPoint pt2, CPoint pt3, CPoint *pPtIntersec)
{
	if (pt2.x == pt1.x || pt2.y == pt1.y)
	{
		if (pPtIntersec != NULL)
		{
			double x = 0, y = 0;
			if (pt2.x == pt1.x)
			{
				x = pt1.x;
				y = pt3.y;
			}
			else if (pt2.y == pt1.y)
			{
				x = pt3.x;
				y = pt1.y;
			}
			pPtIntersec->x = (int)RoundH(x);
			pPtIntersec->y = (int)RoundH(y);
		}

		if (pt2.x == pt1.x && (pt3.y >= ov_min(pt1.y, pt2.y) && pt3.y <= ov_max(pt1.y, pt2.y)))
			return fabs((double)(pt3.x - pt1.x));
		if (pt2.y == pt1.y && (pt3.x >= ov_min(pt1.x, pt2.x) && pt3.x <= ov_max(pt1.x, pt2.x)))
			return fabs((double)(pt3.y - pt1.y));

		return ov_min(GetLineDist(pt1.x, pt1.y, pt3.x, pt3.y), GetLineDist(pt2.x, pt2.y, pt3.x, pt3.y));
	}

	double b1 = (pt2.y - pt1.y)*1.0 / (pt2.x - pt1.x);
	double b2 = -1 / b1;
	double a1 = pt1.y - b1*pt1.x;
	double a2 = pt3.y - b2*pt3.x;

	double x = (a1 - a2) / (b2 - b1);
	double y = a2 + b2*x;

	if (pPtIntersec != NULL)
	{
		pPtIntersec->x = (int)RoundH(x);
		pPtIntersec->y = (int)RoundH(y);
	}

	if (x<ov_min(pt1.x, pt2.x) || x>ov_max(pt1.x, pt2.x) || y<ov_min(pt1.y, pt2.y) || y>ov_max(pt1.y, pt2.y))
	{
		return ov_min(GetLineDist(pt1.x, pt1.y, pt3.x, pt3.y), GetLineDist(pt2.x, pt2.y, pt3.x, pt3.y));
	}
	return GetLineDist(pt3.x, pt3.y, x, y);
}

double GetPoint64ToLineDist(CPoint64 pt1, CPoint64 pt2, CPoint64 pt3, CPoint64 *pPtIntersec)
{
	if (pt2.x == pt1.x || pt2.y == pt1.y)
	{
		if (pPtIntersec != NULL)
		{
			double x = 0, y = 0;
			if (pt2.x == pt1.x)
			{
				x = (double)pt1.x;
				y = (double)pt3.y;
			}
			else if (pt2.y == pt1.y)
			{
				x = (double)pt3.x;
				y = (double)pt1.y;
			}
			pPtIntersec->x = RoundI64(x);
			pPtIntersec->y = RoundI64(y);
		}

		if (pt2.x == pt1.x && (pt3.y >= ov_min(pt1.y, pt2.y) && pt3.y <= ov_max(pt1.y, pt2.y)))
			return fabs((double)(pt3.x - pt1.x));
		if (pt2.y == pt1.y && (pt3.x >= ov_min(pt1.x, pt2.x) && pt3.x <= ov_max(pt1.x, pt2.x)))
			return fabs((double)(pt3.y - pt1.y));

		return ov_min(GetLineDist((double)pt1.x, (double)pt1.y, (double)pt3.x, (double)pt3.y), GetLineDist((double)pt2.x, (double)pt2.y, (double)pt3.x, (double)pt3.y));
	}

	double b1 = (pt2.y - pt1.y)*1.0 / (pt2.x - pt1.x);
	double b2 = -1 / b1;
	double a1 = pt1.y - b1*pt1.x;
	double a2 = pt3.y - b2*pt3.x;

	double x = (a1 - a2) / (b2 - b1);
	double y = a2 + b2*x;

	if (pPtIntersec != NULL)
	{
		pPtIntersec->x = RoundI64(x);
		pPtIntersec->y = RoundI64(y);
	}

	if (x<ov_min(pt1.x, pt2.x) || x>ov_max(pt1.x, pt2.x) || y<ov_min(pt1.y, pt2.y) || y>ov_max(pt1.y, pt2.y))
	{
		return ov_min(GetLineDist((double)pt1.x, (double)pt1.y, (double)pt3.x, (double)pt3.y), GetLineDist((double)pt2.x, (double)pt2.y, (double)pt3.x, (double)pt3.y));
	}
	return GetLineDist((double)pt3.x, (double)pt3.y, x, y);
}

//ԲԲġ͹ȣȡʼǶֹǶ
void GetCadBulgeAngle(double xCenter,double yCenter,double x1,double y1,double x2,double y2,double dBulge,double& dStartAngle,double& dEndAngle)
{
	RVector vv(xCenter,-yCenter);
	RVector v1(x1,-y1);
	RVector v2(x2,-y2);
	dStartAngle=180-vv.getAngleTo(v1)/M_PI*180;
	dEndAngle=180-vv.getAngleTo(v2)/M_PI*180;
	if(dStartAngle<0)
		dStartAngle+=360;
	if(dEndAngle<0)
		dEndAngle+=360;
	if(dBulge>0)
	{
		double tmp=dStartAngle;
		dStartAngle=dEndAngle;
		dEndAngle=tmp;
	}
	if(dStartAngle>dEndAngle)
		dEndAngle+=360;
}

double GetCadBulgeFromMapPoint(double cx, double cy, double x1, double y1, double x2, double y2, bool bClockwise)
{
	// cx, cy: Բ
	// x1, y1, x2, y2: Բϵĵ[x, y]
	// bClockwise: ˳ʱĻ

	// Bulge<0: ˳ʱ, Bulge>0: ʱ

	double dStartRad = 0, dEndRad = 0;

	dStartRad = atan2(-(y1 - cy), -(x1 - cx));
	dEndRad = atan2(-(y2 - cy), -(x2 - cx));

	// double dStartDeg = TO_DEG(dStartRad);
	// double dEndDeg = TO_DEG(dEndRad);

	RVector cc(cx, cy, 0);
	double rr = sqrt(pow(x1 - cx, 2) + pow(y1 - cy, 2));

	double dSign = -1;
	if (!bClockwise)
	{
		double dd = dStartRad;
		dStartRad = dEndRad;
		dEndRad = dd;
		dSign = 1;
	}

	RArc arc(cc, rr, dStartRad, dEndRad);

	return dSign*arc.getBulge();

}

int PtToArcWidth(CRect64 reArea, INT64 x, INT64 y, INT64 iArcX1, INT64 iArcY1, INT64 iArcX2, INT64 iArcY2, double dBulge)
{
	INT64 xmid = reArea.left + reArea.Width() / 2;
	INT64 ymid = reArea.top + reArea.Height() / 2;
	double dAngle = GetLineAngle((double)xmid, (double)ymid, (double)x, (double)y);
	//double dAngle=atan2((ymid-y)*1.0,(xmid-x)*1.0)*180/M_PI;
	//if(dAngle<0)
	//	dAngle+=360;

	double dStartAngle = GetLineAngle((double)xmid, (double)ymid, (double)iArcX1, (double)iArcY1);
	double dEndAngle = GetLineAngle((double)xmid, (double)ymid, (double)iArcX2, (double)iArcY2);
	if(dBulge>0)
	{
		double tmp=dStartAngle;
		dStartAngle=dEndAngle;
		dEndAngle=tmp;
	}
	if(dStartAngle>dEndAngle)
		dEndAngle+=360;

	double dRadius=reArea.Width()/2.0;
	
	if((dAngle>=dStartAngle && dAngle<=dEndAngle) || (dAngle+360>=dStartAngle && dAngle+360<=dEndAngle))
	{
		double aa = GetLineDist((double)xmid, (double)ymid, (double)x, (double)y);
		return (int)RoundH(fabs(aa-dRadius));
	}

	// modify by houzz at 9:56 2016/10/19
	// ,
	// һ. 1.0ǰ
	// . Xֵûг1.0пܻ
	// double dmin=ov_min(sqrt((x-iArcX1)*(x-iArcX1)+(y-iArcY1)*(y-iArcY1)*1.0),sqrt((x-iArcX2)*(x-iArcX2)+(y-iArcY2)*(y-iArcY2)*1.0));
	double dmin = ov_min(sqrt(1.0*(x - iArcX1)*(x - iArcX1) + 1.0*(y - iArcY1)*(y - iArcY1)), sqrt(1.0*(x - iArcX2)*(x - iArcX2) + 1.0*(y - iArcY2)*(y - iArcY2)));

	return (int)RoundH(dmin);
}

int PtToArcWidth(CRect64 reArea, INT64 x, INT64 y, double dStartAngle, double dEndAngle)
{
	INT64 xmid = reArea.left + reArea.Width() / 2;
	INT64 ymid = reArea.top + reArea.Height() / 2;
	double dAngle = GetLineAngle((double)xmid, (double)ymid, (double)x, (double)y);
	double dRadius=reArea.Width()/2.0;
	
	if((dAngle>=dStartAngle && dAngle<=dEndAngle) || (dAngle+360>=dStartAngle && dAngle+360<=dEndAngle))
	{
		double aa = GetLineDist((double)xmid, (double)ymid, (double)x, (double)y);
		return (int)RoundH(fabs(aa-dRadius));
	}

	double iArcX1,iArcX2,iArcY1,iArcY2;
	GetXyInEllipse64(reArea,dStartAngle,iArcX1,iArcY1);
	GetXyInEllipse64(reArea, dEndAngle, iArcX2, iArcY2);

	// ͬ: PtToArcWidth
	// double dmin=ov_min(sqrt((x-iArcX1)*(x-iArcX1)+(y-iArcY1)*(y-iArcY1)*1.0),sqrt((x-iArcX2)*(x-iArcX2)+(y-iArcY2)*(y-iArcY2)*1.0));
	double dmin = ov_min(sqrt(1.0*(x - iArcX1)*(x - iArcX1) + 1.0*(y - iArcY1)*(y - iArcY1)), sqrt(1.0*(x - iArcX2)*(x - iArcX2) + 1.0*(y - iArcY2)*(y - iArcY2)));

	return (int)RoundH(dmin);
}



void RectAddPoint64(RECTH64& re, INT64 x, INT64 y)
{
	if (re.left == -1 && re.top == -1 && re.right == -1 && re.bottom == -1)
	{
		if (x == -1 && y == -1)
		{
			re.right = 0;
			re.bottom = 0;
		}
		else
		{
			re.left = re.right = x;
			re.top = re.bottom = y;
		}
		return;
	}

	if (x<re.left)
		re.left = x;
	if (x>re.right)
		re.right = x;
	if (y<re.top)
		re.top = y;
	if (y>re.bottom)
		re.bottom = y;
}

void RectAddPoint(RECT& re,int x,int y)
{
	if(re.left==-1 && re.top==-1 && re.right==-1 && re.bottom==-1)
	{
		if (x == -1 && y == -1)
		{
			re.right = 0;
			re.bottom = 0;
		}
		else
		{
			re.left = re.right = x;
			re.top = re.bottom = y;
		}
		return;
	}

	if(x<re.left)
		re.left=x;
	if(x>re.right)
		re.right=x;
	if(y<re.top)
		re.top=y;
	if(y>re.bottom)
		re.bottom=y;
}

void RectAddSize64(RECTH64& re, INT64 cx, INT64 cy)
{
	re.left -= cx;
	re.right += cx;
	re.top -= cy;
	re.bottom += cy;
}

void RectAddSize(RECT& re, int cx, int cy)
{
	re.left -= cx;
	re.right += cx;
	re.top -= cy;
	re.bottom += cy;
}

CRect64 GetArcDrawArea64(CRect64 reArc, double dStartAngle, double dEndAngle)
{
	if (dEndAngle - dStartAngle>180)
	{
		return reArc;
	}

	double x1, y1, x2, y2;
	GetXyInEllipse64(reArc, dStartAngle, x1, y1);
	GetXyInEllipse64(reArc, dEndAngle, x2, y2);
	RECTH64 reDraw = CRect64((INT64)x1, (INT64)y1, (INT64)x1, (INT64)y1);
	RectAddPoint64(reDraw, (INT64)x2, (INT64)y2);

	if (dStartAngle<360 && dEndAngle>360)
		RectAddPoint64(reDraw, reArc.left, (reArc.top + reArc.bottom) / 2);
	if ((dStartAngle<90 && dEndAngle>90) || (dStartAngle<450 && dEndAngle>450))
		RectAddPoint64(reDraw, (reArc.left + reArc.right) / 2, reArc.top);
	if (dStartAngle<180 && dEndAngle>180)
		RectAddPoint64(reDraw, reArc.right, (reArc.top + reArc.bottom) / 2);
	if (dStartAngle<270 && dEndAngle>270)
		RectAddPoint64(reDraw, (reArc.left + reArc.right) / 2, reArc.bottom);


	return reDraw;
}

CRect GetArcDrawArea(CRect reArc,double dStartAngle,double dEndAngle)
{
	CRect64 reArc64 = Rect32To64(reArc);
	CRect64 reDraw64 = GetArcDrawArea64(reArc64, dStartAngle, dEndAngle);
	return Rect64To32(reDraw64);

	if(dEndAngle-dStartAngle>180)
	{
		return reArc;
	}

	double x1,y1,x2,y2;
	GetXyInEllipse(reArc,dStartAngle,x1,y1);
	GetXyInEllipse(reArc,dEndAngle,x2,y2);
	RECT reDraw=CRect((int)x1,(int)y1,(int)x1,(int)y1);
	RectAddPoint(reDraw,(int)x2,(int)y2);

	if(dStartAngle<360 && dEndAngle>360)
		RectAddPoint(reDraw,reArc.left,(reArc.top+reArc.bottom)/2);
	if((dStartAngle<90 && dEndAngle>90) || (dStartAngle<450 && dEndAngle>450))
		RectAddPoint(reDraw,(reArc.left+reArc.right)/2,reArc.top);
	if(dStartAngle<180 && dEndAngle>180)
		RectAddPoint(reDraw,reArc.right,(reArc.top+reArc.bottom)/2);
	if(dStartAngle<270 && dEndAngle>270)
		RectAddPoint(reDraw,(reArc.left+reArc.right)/2,reArc.bottom);


	return reDraw;
}


void GetEdushiOffset(double angle,double scale,double x,double y,int& xx,int& yy)
{
	if(angle>0)
	{
		x*=scale;
		y*=scale;
	}
	else
	{
		x/=scale;
		y/=scale;
	}
    
	double fSide = sqrt(x*x+y*y);
	double angleA1 = atan2(y, x);
	double angleA2 = angleA1+2*M_PI*angle/360;
	yy = (int)(fSide*sin(angleA2));
	xx = (int)(fSide*cos(angleA2));
}

void GetEdushiOffset(POINTH64 ptSrc1, POINTH64 ptSrc2, POINTH64 ptSrc3, POINTH64 ptDst1, POINTH64 ptDst2, POINTH64 &ptDst3)
{
	if (ptSrc3.x == ptSrc1.x && ptSrc3.y == ptSrc1.y)
	{
		ptDst3 = ptDst1;
		return;
	}
    
	double Src_12_x = (double)(ptSrc2.x-ptSrc1.x);
	double Src_12_y = (double)(ptSrc2.y-ptSrc1.y);
    
	double Src_13_x = (double)(ptSrc3.x-ptSrc1.x);
	double Src_13_y = (double)(ptSrc3.y-ptSrc1.y);
    
	double Dst_12_x = (double)(ptDst2.x-ptDst1.x);
	double Dst_12_y = (double)(ptDst2.y-ptDst1.y);
    
	double rakeSrc_12 = atan2(Src_12_y, Src_12_x);
	double rakeSrc_13 = atan2(Src_13_y, Src_13_x);
    
	double angleSrc_12_13 = rakeSrc_12-rakeSrc_13;
    
	double rakeDst_12 = atan2(Dst_12_y, Dst_12_x);
	double rakeDst_13 = rakeDst_12-angleSrc_12_13;
    
	double SrcLen_12 = sqrt(Src_12_x*Src_12_x+Src_12_y*Src_12_y);
	double SrcLen_13 = sqrt(Src_13_x*Src_13_x+Src_13_y*Src_13_y);
	double DstLen_12 = sqrt(Dst_12_x*Dst_12_x+Dst_12_y*Dst_12_y);
    
	double DstLen_13 = DstLen_12*SrcLen_13/SrcLen_12;
    
    
	double yy = ((double)ptDst1.y+DstLen_13*sin(rakeDst_13));
	double xx = ((double)ptDst1.x+DstLen_13*cos(rakeDst_13));
	ptDst3.y = (INT64)yy;
	ptDst3.x = (INT64)xx;
    
    
	if (fabs((double)ptDst3.y-yy+1)<1.0e-6)
		ptDst3.y += 1;
	if (fabs((double)ptDst3.y-yy-1)<1.0e-6)
		ptDst3.y -= 1;
	if (fabs((double)ptDst3.x-xx+1)<1.0e-6)
		ptDst3.x += 1;
	if (fabs((double)ptDst3.x-xx-1)<1.0e-6)
		ptDst3.x -= 1;
}

#define ABS_64(x) ((x)<0?-(x):(x))
//ֵ1, 2, 3, 00ʾڸ
int NearlyPos(POINTH64 ptSrc1, POINTH64 ptSrc2, POINTH64 ptSrc3, POINTH64 ptSrc4, int nOffset)
{
	if (ABS_64(ptSrc4.x-ptSrc1.x)<=nOffset && ABS_64(ptSrc4.y-ptSrc1.y)<=nOffset)
		return 1;
	else if (ABS_64(ptSrc4.x-ptSrc2.x)<=nOffset && ABS_64(ptSrc4.y-ptSrc2.y)<=nOffset)
		return 2;
	else if (ABS_64(ptSrc4.x-ptSrc3.x)<=nOffset && ABS_64(ptSrc4.y-ptSrc3.y)<=nOffset)
		return 3;
	return 0;
}

double GetPointDis(POINT u, POINT v)
{
	return sqrt((double)((u.x - v.x)*(u.x - v.x) + (u.y - v.y)*(u.y - v.y)));
}

//ֱƽ
bool IsLineParallel(POINT u1, POINT u2, POINT v1, POINT v2)
{
	return (u1.x - u2.x)*(v1.y - v2.y) - (v1.x - v2.x)*(u1.y - u2.y) == 0;
}

bool TwoLineInteractionRv(RVector point1, RVector point2, RVector point3, RVector point4, RVector& pointInterSection)
{
	double d = (point2.y - point1.y) * (point3.x - point4.x) - (point2.x - point1.x) * (point3.y - point4.y);
	if (d == 0)
		return false;

	double t1 = ((point3.y - point4.y) * (point1.x - point3.x) - (point3.x - point4.x) * (point1.y - point3.y))
		/ d;
	double t2 = ((point2.x - point1.x) * (point1.y - point3.y) - (point2.y - point1.y) * (point1.x - point3.x))
		/ d;

	// ʽ
	pointInterSection.x = point1.x + (point2.x - point1.x) * t1;
	pointInterSection.y = point1.y + (point2.y - point1.y) * t1;

	return (t1 >= 0 && t1 <= 1 && t2 >= 0 && t2 <= 1);
}

bool GetLineRectInteraction(RVector pt1, RVector pt2, CRect re, RVector* pRvInt, int* pnRvInt)
{
	bool bInt = false;
	RVector rvInt;
	int& nRvInt = *pnRvInt;
	if (TwoLineInteractionRv(pt1, pt2, RVector(re.left, re.top), RVector(re.left, re.bottom), rvInt))
	{
		if (pRvInt)
		{
			pRvInt[nRvInt] = rvInt;
			nRvInt++;
		}
		else
			return true;
		bInt = true;
	}
	if (TwoLineInteractionRv(pt1, pt2, RVector(re.left, re.top), RVector(re.right, re.top), rvInt))
	{
		if (pRvInt)
		{
			pRvInt[nRvInt] = rvInt;
			nRvInt++;
		}
		else
			return true;
		bInt = true;
	}
	if (TwoLineInteractionRv(pt1, pt2, RVector(re.left, re.bottom), RVector(re.right, re.bottom), rvInt))
	{
		if (pRvInt)
		{
			pRvInt[nRvInt] = rvInt;
			nRvInt++;
		}
		else
			return true;
		bInt = true;
	}
	if (TwoLineInteractionRv(pt1, pt2, RVector(re.right, re.top), RVector(re.right, re.bottom), rvInt))
	{
		if (pRvInt)
		{
			pRvInt[nRvInt] = rvInt;
			nRvInt++;
		}
		else
			return true;
		bInt = true;
	}
	if (bInt)
		return true;

	//ûڲ
	if (re.PtInRect(CPoint((int)pt1.x, (int)pt1.y)))
		return true;
	if (re.PtInRect(CPoint((int)pt2.x, (int)pt2.y)))
		return true;
	return false;
}

bool GetLineRectInteraction64(RVector pt1, RVector pt2, CRect64 re, RVector* pRvInt, int* pnRvInt)
{
	bool bInt = false;
	RVector rvInt;
	int& nRvInt = *pnRvInt;
	if (TwoLineInteractionRv(pt1, pt2, RVector(1.0*re.left, 1.0*re.top), RVector(1.0*re.left, 1.0*re.bottom), rvInt))
	{
		if (pRvInt)
		{
			pRvInt[nRvInt] = rvInt;
			nRvInt++;
		}
		else
			return true;
		bInt = true;
	}
	if (TwoLineInteractionRv(pt1, pt2, RVector(1.0*re.left, 1.0*re.top), RVector(1.0*re.right, 1.0*re.top), rvInt))
	{
		if (pRvInt)
		{
			pRvInt[nRvInt] = rvInt;
			nRvInt++;
		}
		else
			return true;
		bInt = true;
	}
	if (TwoLineInteractionRv(pt1, pt2, RVector(1.0*re.left, 1.0*re.bottom), RVector(1.0*re.right, 1.0*re.bottom), rvInt))
	{
		if (pRvInt)
		{
			pRvInt[nRvInt] = rvInt;
			nRvInt++;
		}
		else
			return true;
		bInt = true;
	}
	if (TwoLineInteractionRv(pt1, pt2, RVector(1.0*re.right, 1.0*re.top), RVector(1.0*re.right, 1.0*re.bottom), rvInt))
	{
		if (pRvInt)
		{
			pRvInt[nRvInt] = rvInt;
			nRvInt++;
		}
		else
			return true;
		bInt = true;
	}
	if (bInt)
		return true;

	//ûڲ
	if (re.PtInRect(CPoint64((INT64)pt1.x, (INT64)pt1.y)))
		return true;
	if (re.PtInRect(CPoint64((INT64)pt2.x, (INT64)pt2.y)))
		return true;
	return false;
}

//pt3 ֱpt1->pt2Ľ
RVector GetPointToLineInteraction(RVector pt1, RVector pt2, RVector pt3)
{
	RVector rvInt(0, 0);
	if (pt2.x == pt1.x || pt2.y == pt1.y)
	{
		if (pt2.x == pt1.x)
		{
			rvInt.x = pt1.x;
			rvInt.y = pt3.y;
		}
		else if (pt2.y == pt1.y)
		{
			rvInt.x = pt3.x;
			rvInt.y = pt1.y;
		}
		return rvInt;
	}

	double b1 = (pt2.y - pt1.y) / (pt2.x - pt1.x);
	double b2 = -1 / b1;
	double a1 = pt1.y - b1*pt1.x;
	double a2 = pt3.y - b2*pt3.x;

	rvInt.x = (a1 - a2) / (b2 - b1);
	rvInt.y = a2 + b2*rvInt.x;
	return rvInt;
}

//ȡ(1234)
bool TwoLineInteraction(POINT point1, POINT point2, POINT point3, POINT point4, POINT& pointInterSection)
{
#if 1
	// t1<0intersectionpoint1֮,t1>1intersectionpoint2֮
	// ͬt2


	double d = (double)(point2.y - point1.y) * (double)(point3.x - point4.x) - (double)(point2.x - point1.x) * (double)(point3.y - point4.y);
	if (d == 0)
		return false;

	double t1 = ((double)(point3.y - point4.y) * (double)(point1.x - point3.x) - (double)(point3.x - point4.x) * (double)(point1.y - point3.y))
		/ d;
	double t2 = ((double)(point2.x - point1.x) * (double)(point1.y - point3.y) - (double)(point2.y - point1.y) * (double)(point1.x - point3.x))
		/ d;
	// ʽ
	pointInterSection.x = (int)((double)(point1.x) + (double)(point2.x - point1.x) * t1);
	pointInterSection.y = (int)((double)(point1.y) + (double)(point2.y - point1.y) * t1);

#ifdef _DEBUG
	int xx = (int)((double)(point3.x) + (double)(point4.x - point3.x) * t2);
	int yy = (int)((double)(point3.y) + (double)(point4.y - point3.y) * t2);
	if (pointInterSection.x != xx || pointInterSection.y != yy){
		yy = yy;
	}
#endif

	return (t1 >= 0 && t1 <= 1 && t2 >= 0 && t2 <= 1);
#else
	//kΪintersection()point1ľpoint1point2ľ,k<0intersectionpoint1֮,k>1intersectionpoint2֮
	double k = ((double)(point1.y - point4.y) * (double)(point3.x - point4.x)- (double)(point1.x - point4.x) * (double)(point3.y - point4.y))
		/((double)(point2.x - point1.x) * (double)(point3.y - point4.y) - (double)(point2.y - point1.y) * (double)(point3.x - point4.x));
	pointInterSection.x = (int)((double)(point1.x) + (double)(point2.x - point1.x) * k);
	pointInterSection.y = (int)((double)(point1.y) + (double)(point2.y - point1.y) * k);

	return (k >= 0 && k <= 1);
#endif
    
}

//ȡ(1234)
bool TwoLineInteraction64(POINTH64 point1, POINTH64 point2, POINTH64 point3, POINTH64 point4, POINTH64& pointInterSection)
{
	// t1<0intersectionpoint1֮,t1>1intersectionpoint2֮
	// ͬt2

	double d = (double)(point2.y - point1.y) * (double)(point3.x - point4.x) - (double)(point2.x - point1.x) * (double)(point3.y - point4.y);
	if (d == 0)
		return false;
	double t1 = ((double)(point3.y - point4.y) * (double)(point1.x - point3.x) - (double)(point3.x - point4.x) * (double)(point1.y - point3.y))
		/ d;
	double t2 = ((double)(point2.x - point1.x) * (double)(point1.y - point3.y) - (double)(point2.y - point1.y) * (double)(point1.x - point3.x))
		/ d;
	// ʽ
	pointInterSection.x = (INT64)((double)(point1.x) + (double)(point2.x - point1.x) * t1);
	pointInterSection.y = (INT64)((double)(point1.y) + (double)(point2.y - point1.y) * t1);

	return (t1 >= 0 && t1 <= 1 && t2 >= 0 && t2 <= 1);
}

void GetPlaneEdushiOffset(POINTH64 ptSrc1, POINTH64 ptSrc2, POINTH64 ptSrc3, POINTH64 ptSrc4, POINTH64 ptDst1, POINTH64 ptDst2, POINTH64 ptDst3, POINTH64 &ptDst4)
{
    
	POINTH64 ptSrcArray[3] = { ptSrc1, ptSrc2, ptSrc3 };
	POINTH64 ptDstArray[3] = { ptDst1, ptDst2, ptDst3 };
    
	int nNear = NearlyPos(ptSrc1, ptSrc2, ptSrc3, ptSrc4, 1);
	if (nNear != 0)
	{
		ptDst4 = ptDstArray[nNear-1];
		return;
	}
    
	int nVertex = 1;
	double dRakeMax = 0;
    
	int i;
	// жĸб
	for (i=1; i<=3; i++)
	{
		// ֱߵļнһСڻڶ(PI/2)
		double dRake1 = atan2((double)(ptSrc4.y-ptSrcArray[i-1].y), (double)(ptSrc4.x-ptSrcArray[i-1].x));
		double dRake2 = atan2((double)(ptSrcArray[i%3].y-ptSrcArray[(i+1)%3].y), (double)(ptSrcArray[i%3].x-ptSrcArray[(i+1)%3].x));
        
		// ֵڷΧ(-2PI~+2PI)Ҫת(0~PI/2)
		double dd = dRake1-dRake2;
		if (dd<0)
			dd = -dd;
        
		if (dd > M_PI && dd < 2*M_PI)
			dd = 2*M_PI - dd;
		else if (dd*2 > M_PI)
			dd = M_PI - dd;
        
		if (dRakeMax<dd)
		{
			dRakeMax = dd;
			nVertex = i;
		}
	}
    
	POINTH64 pointInterSrc = {0};
	TwoLineInteraction64(ptSrc4, ptSrcArray[nVertex - 1], ptSrcArray[(nVertex) % 3], ptSrcArray[(nVertex + 1) % 3], pointInterSrc);
    
    
	POINTH64 pointInterDst = {0};
	// 󽻵pointInterSrcһƽжӦĵpointInterDst
	GetEdushiOffset(ptSrcArray[(nVertex)%3], ptSrcArray[(nVertex+1)%3], pointInterSrc,ptDstArray[(nVertex)%3], ptDstArray[(nVertex+1)%3], pointInterDst);
    
	// ptSrc4һƽжӦĵptDst4
	GetEdushiOffset(pointInterSrc, ptSrcArray[nVertex-1], ptSrc4, pointInterDst, ptDstArray[nVertex-1], ptDst4);
    
	return;
}

bool IsSameSlope(POINT pt1, POINT pt2, POINT pt3)
{
	double s1 = atan2((double)pt3.y - pt2.y, pt3.x - pt2.x);
	double s2 = atan2((double)pt2.y - pt1.y, pt2.x - pt1.x);
	return fabs(s1 - s2) < M_PI / 180;
}

CRect64 GetEllipseDrawBox(RVector rvCenter, RVector rvEnd, double dAxisScale, int iWidth, double dStartAngle, double dEndAngle)
{
	double aAxis = rvEnd.getDistanceTo(rvCenter);
	double bAxis = aAxis*dAxisScale;
	double dMaxAxis = ov_max(aAxis, bAxis);
	CRect64 reEllipseBox((int)(rvCenter.x - dMaxAxis), (int)(rvCenter.y - dMaxAxis), (int)(rvCenter.x + dMaxAxis), (int)(rvCenter.y + dMaxAxis));

	if (dEndAngle - dStartAngle < 90)
	{
		double dxStartP, dyStartP, dxEndP, dyEndP;
		GetXyInTiltEllipse(rvCenter, rvEnd, dAxisScale, dStartAngle, dxStartP, dyStartP);
		GetXyInTiltEllipse(rvCenter, rvEnd, dAxisScale, dEndAngle, dxEndP, dyEndP);

		RVector rvStart(dxStartP, dyStartP);
		RVector rvEnd(dxEndP, dyEndP);
		RVector rvM = rvStart;
		rvM.scale(0.5, rvEnd);
		double dRadius = rvM.getDistanceTo(rvEnd) + iWidth + 2;

		CRect64 reBox(-1, -1, -1, -1);
		RectAddPoint64(reBox, (INT64)(rvM.x - dRadius), (INT64)(rvM.y - dRadius));
		RectAddPoint64(reBox, (INT64)(rvM.x + dRadius), (INT64)(rvM.y + dRadius));
		return reBox;
	}
	else
		return reEllipseBox;
}

bool PtInEllipse(RVector rvCenter, RVector rvEnd, double dAxisScale, double x, double y)
{
	if (dAxisScale == 0)
		dAxisScale = 1;
	double aAxis = rvCenter.getDistanceTo(rvEnd);
	double bAxis = aAxis*dAxisScale;

	RVector rv(x, y);

	double angle = rvEnd.getAngleTo(rvCenter);
	rv.rotate(-angle, rvCenter);
	if (aAxis < bAxis)
	{
		double dTmp = aAxis;
		aAxis = bAxis;
		bAxis = dTmp;
		rv.rotate(M_PI / 2, rvCenter);
	}
	rv -= rvCenter;

	if ((rv.x*rv.x) / (aAxis*aAxis) + (rv.y*rv.y) / (bAxis*bAxis) < 1)
		return true;
	else
		return false;
}

bool PtInEllipse(RVector rvCenter, RVector rvEnd, double dAxisScale, double x, double y, double dStartAngle, double dEndAngle)
{
	if (dStartAngle == 0 && dEndAngle == 360)
		return PtInEllipse(rvCenter, rvEnd, dAxisScale, x, y);

	RVector rv(x, y);
	double dAngleRv = rv.getAngleTo(rvCenter);
	double dAngleStart = rvEnd.getAngleTo(rvCenter);
	double dAngle = (dAngleRv - dAngleStart) / M_PI * 180;
	if (dAngle < 0)
		dAngle += 360;

	if ((dAngle >= dStartAngle && dAngle <= dEndAngle) || (dAngle + 360 >= dStartAngle && dAngle + 360 <= dEndAngle))
		return PtInEllipse(rvCenter, rvEnd, dAxisScale, x, y);

	return false;
}

//Xʱ׼Ϊx^2/a^2+y^2/b^2=1 (a>b>0)
// aΪᣬbΪ̰
//Mx0y0 Բ x^2/a^2+y^2/b^2=1 
//Բڣ x0^2/a^2+y0^2/b^2<1 
//Բϣ x0^2/a^2+y0^2/b^2=1 
//Բ⣺ x0^2/a^2+y0^2/b^2>1
bool PtInEllipse(CRect reArea, int x, int y)
{
	// double dx = (reArea.left + reArea.right) / 2.0;
	// double dy = (reArea.top + reArea.bottom) / 2.0;

	int cx = reArea.right - reArea.left - 1;
	int cy = reArea.bottom - reArea.top - 1;
	if (cx <= 0 || cy <= 0)
		return false;

	double xx = x - (reArea.left + cx / 2.0);
	double yy = y - (reArea.top + cy / 2.0);
	if (cx < cy)
	{
		double a = cy / 2.0;
		double b = cx / 2.0;

		if ((xx*xx) / (b*b) + (yy*yy) / (a*a) < 1)
			return true;
		else
			return false;
	}
	else
	{
		double a = cx / 2.0;
		double b = cy / 2.0;

		if ((xx*xx) / (a*a) + (yy*yy) / (b*b) < 1)
			return true;
		else
			return false;
	}
}

bool Pt64InEllipse(CRect64 reArea, INT64 x, INT64 y)
{
	// double dx = (reArea.left + reArea.right) / 2.0;
	// double dy = (reArea.top + reArea.bottom) / 2.0;

	INT64 cx = reArea.right - reArea.left - 1;
	INT64 cy = reArea.bottom - reArea.top - 1;
	if (cx <= 0 || cy <= 0)
		return false;

	double xx = x - (reArea.left + cx / 2.0);
	double yy = y - (reArea.top + cy / 2.0);
	if (cx < cy)
	{
		double a = cy / 2.0;
		double b = cx / 2.0;

		if ((xx*xx) / (b*b) + (yy*yy) / (a*a) < 1)
			return true;
		else
			return false;
	}
	else
	{
		double a = cx / 2.0;
		double b = cy / 2.0;

		if ((xx*xx) / (a*a) + (yy*yy) / (b*b) < 1)
			return true;
		else
			return false;
	}
}

bool PtInSector(RVector rvCenter, RVector rvEnd, double dAxisScale, double x, double y, double dStartAngle, double dEndAngle)
{
	if (dStartAngle == 0 && dEndAngle == 360)
		return PtInEllipse(rvCenter, rvEnd, dAxisScale, x, y);

	double dAngleR = rvEnd.getAngleTo(rvCenter);
	double dAngle = RVector(x,y).getAngleTo(rvCenter) - dAngleR;
	dAngle = StandrandRadian(M_PI + dAngle) * 180 / M_PI;

	if (dAngle >= dStartAngle && dAngle <= dEndAngle)
		return PtInEllipse(rvCenter, rvEnd, dAxisScale, x, y);

	return false;
}

bool PtInSector(CRect reArea, int x, int y, double dStartAngle, double dEndAngle)
{
	if (dStartAngle == 0 && dEndAngle == 360)
		return PtInEllipse(reArea, x, y);

	int xmid = reArea.left + reArea.Width() / 2;
	int ymid = reArea.top + reArea.Height() / 2;
	double dAngle = atan2((ymid - y)*1.0, (xmid - x)*1.0) * 180 / M_PI;
	if (dAngle < 0)
		dAngle += 360;

	if ((dAngle >= dStartAngle && dAngle <= dEndAngle) || (dAngle + 360 >= dStartAngle && dAngle + 360 <= dEndAngle))
		return PtInEllipse(reArea, x, y);

	return false;
}

bool Pt64InSector(CRect64 reArea, INT64 x, INT64 y, double dStartAngle, double dEndAngle)
{
	if (dStartAngle == 0 && dEndAngle == 360)
		return Pt64InEllipse(reArea, x, y);

	INT64 xmid = reArea.left + reArea.Width() / 2;
	INT64 ymid = reArea.top + reArea.Height() / 2;
	double dAngle = atan2((ymid - y)*1.0, (xmid - x)*1.0) * 180 / M_PI;
	if (dAngle < 0)
		dAngle += 360;

	if ((dAngle >= dStartAngle && dAngle <= dEndAngle) || (dAngle + 360 >= dStartAngle && dAngle + 360 <= dEndAngle))
		return Pt64InEllipse(reArea, x, y);

	return false;
}

void GetBoxRotateVector(RVector* prvCenter, double iWidth, double iHeight, double fAngle, RVector* prvLt, RVector* prvRt, RVector* prvLb, RVector* prvRb, RVector* prvLeft, RVector* prvRight, RVector* prvTop, RVector* prvBottom, RVector* prvRotate, int* piRadius)
{
	double x = prvCenter->x;
	double y = prvCenter->y;

	double angle = fAngle / 180.0*M_PI;
	if (prvLt)
	{
		*prvLt = RVector(x - iWidth / 2, y - iHeight / 2);
		prvLt->rotate(angle, *prvCenter);
	}
	if (prvRt)
	{
		*prvRt = RVector(x - iWidth / 2 + iWidth, y - iHeight / 2);
		prvRt->rotate(angle, *prvCenter);
	}
	if (prvLb)
	{
		*prvLb = RVector(x - iWidth / 2, y - iHeight / 2 + iHeight);
		prvLb->rotate(angle, *prvCenter);
	}
	if (prvRb)
	{
		*prvRb = RVector(x - iWidth / 2 + iWidth, y - iHeight / 2 + iHeight);
		prvRb->rotate(angle, *prvCenter);
	}
	if (prvLeft)
	{
		*prvLeft = RVector(x - iWidth / 2, y);
		prvLeft->rotate(angle, *prvCenter);
	}
	if (prvRight)
	{
		*prvRight = RVector(x - iWidth / 2 + iWidth, y);
		prvRight->rotate(angle, *prvCenter);
	}
	if (prvTop)
	{
		*prvTop = RVector(x, y - iHeight / 2);
		prvTop->rotate(angle, *prvCenter);
	}
	if (prvBottom)
	{
		*prvBottom = RVector(x, y - iHeight / 2 + iHeight);
		prvBottom->rotate(angle, *prvCenter);
	}
	if (prvLeft && prvRight && prvRotate)
	{
		RVector v1 = prvLeft->getLerp(*prvRight, 0.11);
		*prvRotate = prvLeft->getLerp(*prvRight, 0.14);
		if (piRadius)
		{
			*piRadius = ov_max(3, (int)RoundH(v1.getDistanceTo(*prvRotate)));
			if (*piRadius > 10)
				*piRadius = 10;
		}
	}
}

void RVectorRotate(RVector rvCenter, RVector& rv, double fAngle)
{
	double angle = -fAngle / 180.0*M_PI;
	rv.rotate(angle, rvCenter);
}

CRect GetRectRotate(CRect re, RVector rvCenter, double fAngle, RVector* prv4)
{
	CRect64 re1(re.left, re.top, re.right, re.bottom);
	CRect64 re2 = GetRectRotate64(re1, rvCenter, fAngle, prv4);
	CRect reRet((int)re2.left, (int)re2.top, (int)re2.right, (int)re2.bottom);
	return reRet;
	
}

CRect64 GetRectRotate64(CRect64 re, RVector rvCenter, double fAngle, RVector* prv4)
{
	RVector rvLt((double)re.left, (double)re.top);
	RVector rvRt((double)re.right, (double)re.top);
	RVector rvLb((double)re.left, (double)re.bottom);
	RVector rvRb((double)re.right, (double)re.bottom);
	double angle = -fAngle / 180.0*M_PI;
	rvLt.rotate(angle, rvCenter);
	rvRt.rotate(angle, rvCenter);
	rvLb.rotate(angle, rvCenter);
	rvRb.rotate(angle, rvCenter);

	CRect64 reBox(-1, -1, -1, -1);
	RectAddPoint64(reBox, RoundI64(rvLt.x), RoundI64(rvLt.y));
	RectAddPoint64(reBox, RoundI64(rvLb.x), RoundI64(rvLb.y));
	RectAddPoint64(reBox, RoundI64(rvRt.x), RoundI64(rvRt.y));
	RectAddPoint64(reBox, RoundI64(rvRb.x), RoundI64(rvRb.y));

	if (prv4)
	{
		prv4[0] = rvLt;
		prv4[1] = rvRt;
		prv4[2] = rvLb;
		prv4[3] = rvRb;
	}
	return reBox;
}


CRect GetRectRotateLb(CRect re, double fAngle, RVector* prv4)
{
	CRect64 re1(re.left, re.top, re.right, re.bottom);
	CRect64 re2 = GetRectRotateLb64(re1, fAngle, prv4);
	CRect reRet((int)re2.left, (int)re2.top, (int)re2.right, (int)re2.bottom);
	return reRet;
}

CRect64 GetRectRotateLb64(CRect64 re, double fAngle, RVector* prv4)
{
	RVector rvCenter((double)re.left, (double)re.bottom);
	return GetRectRotate64(re, rvCenter, fAngle, prv4);

	RVector rvLt, rvRt, rvLb, rvRb;
	GetBoxRotateVector(&rvCenter, (double)re.Width(), (double)re.Height(), fAngle, &rvLt, &rvRt, &rvLb, &rvRb, NULL, NULL, NULL, NULL, NULL, NULL);
	CRect64 reBox(-1, -1, -1, -1);
	RectAddPoint64(reBox, RoundI64(rvLt.x), RoundI64(rvLt.y));
	RectAddPoint64(reBox, RoundI64(rvLb.x), RoundI64(rvLb.y));
	RectAddPoint64(reBox, RoundI64(rvRt.x), RoundI64(rvRt.y));
	RectAddPoint64(reBox, RoundI64(rvRb.x), RoundI64(rvRb.y));

	if (prv4)
	{
		prv4[0] = rvLt;
		prv4[1] = rvRt;
		prv4[2] = rvLb;
		prv4[3] = rvRb;
	}

	RVector rvMid(re.left + re.Width() / 2.0, re.top + re.Height() / 2.0);
	RVectorRotate(rvCenter, rvMid, fAngle);
	//double angle = -fAngle / 180.0*M_PI;
	//reMid.rotate(angle, rvCenter);


	double xTmp = rvMid.x - reBox.Width() / 2.0;
	double yTmp = rvMid.y - reBox.Height() / 2.0;
	CRect64 reNew;
	reNew.left = RoundI64(xTmp);
	reNew.top = RoundI64(yTmp);
	reNew.right = reNew.left + reBox.Width();
	reNew.bottom = reNew.top + reBox.Height();

	return reNew;
}

//ֵ
//2DֱԲĽ
int GetLineInteractionCircle(RVector rv1, RVector rv2, RVector rvCenter, double dRadius, RVector& rvInt1, RVector& rvInt2)
{
	rvInt1.x = rvInt1.y = rvInt1.z = 0;
	rvInt2.x = rvInt2.y = rvInt2.z = 0;

	if (dRadius < DOUBLE_EQ_DIFF)
		return 0;

	RVector rvInt = GetPointToLineInteraction(rv1, rv2, rvCenter);
	double d1 = rvInt.getDistanceTo(rvCenter);
	if (fabs(d1 - dRadius) < DOUBLE_EQ_DIFF)
	{
		//
		rvInt1 = rvInt;
		return 1;
	}
	else if (d1 > dRadius)
	{
		//޽
		return 0;
	}
	else if (d1 < DOUBLE_EQ_DIFF)
	{
		//ྭԲ
		rvInt1 = rv1;
		double ds = dRadius / rv1.getDistanceTo(rvCenter);
		rvInt1.scale(ds, rvCenter);
		rvInt2 = rv1;
		rvInt2.scale(-ds, rvCenter);
		return 2;
	}

	// rv1rv2Ĵֱ  Բ㵽 ļн
	double da = M_PI/2 - asin(d1 / dRadius);

	// rv1rv2ĴֱԲĽ
	RVector rv = rvInt;
	rv.scale(dRadius/d1, rvCenter);

	rvInt1 = rv;
	rvInt1.rotate(da, rvCenter);
	rvInt2 = rv;
	rvInt2.rotate(-da, rvCenter);
	return 2;
}

//Ƕ:ҶCosA=(b^2+c^2-a^2)/(2bc)
double GetTriangleAngleByEdgeSize(double dAngleEdge, double dEdge1, double dEdge2)
{
	if ((dAngleEdge - dEdge1 - dEdge2) > 0.00000001)
		return M_PI;
	if ((dEdge1 - dAngleEdge - dEdge2) > 0.00000001 || (dEdge2 - dAngleEdge - dEdge1) > 0.00000001)
		return 0;

	double dd = (dEdge1*dEdge1 + dEdge2*dEdge2 - dAngleEdge*dAngleEdge) / (2 * dEdge1*dEdge2);
	return acos(dd);
}

//ԲཻĽ
int TwoCircleInteraction(RVector rv1, double dRadius1, RVector rv2, double dRadius2, RVector& rvInt1, RVector& rvInt2)
{
	rvInt1.x = rvInt1.y = rvInt1.z = 0;
	rvInt2.x = rvInt2.y = rvInt2.z = 0;
	if (dRadius1 < DOUBLE_EQ_DIFF || dRadius2 < DOUBLE_EQ_DIFF)
		return 0;
	double dd = rv1.getDistanceTo(rv2);
	if (dd < DOUBLE_EQ_DIFF)
		return 0;
	else if (fabs(dd - dRadius1 - dRadius2) < DOUBLE_EQ_DIFF)
	{
		//
		rvInt1 = rv1;
		double ds = dRadius2 / dd;
		rvInt1.scale(ds, rv2);
		return 1;
	}

	//Ƕ:ҶCosA=(b^2+c^2-a^2)/(2bc)
	//double da = acos((dd*dd + dRadius1*dRadius1 - dRadius2*dRadius2) / (2 * dd*dRadius1));
	double da = GetTriangleAngleByEdgeSize(dRadius2, dd, dRadius1);

	//rv1-rv2Բ1
	RVector rv = rv2;
	double ds = dRadius1 / dd;
	rv.scale(ds, rv1);

	rvInt1 = rv;
	rvInt1.rotate(da, rv1);
	rvInt2 = rv;
	rvInt2.rotate(-da, rv1);
	return 2;
}

//D3DXIntersectTri
//rvOrgΪԴʼ
//ע⣺rvDirΪrvDir=rvDst-rvOrg;
//rv0rv1rv2Ϊ㣬bPlaneCheckOnlyʾǷֻƽûнжϽû
//ο㷨Դhttp://blog.csdn.net/heyzol/article/details/17348317
bool IntersectTriangle(RVector rvOrg, RVector rvDir, RVector rv0, RVector rv1, RVector rv2, bool bPlaneCheckOnly, double *t, double *u, double *v)
{
	*t = *u = *v = 0;
	RVector rvE1 = rv1 - rv0;
	RVector rvE2 = rv2 - rv0;

	RVector rvP = rvDir.Cross(rvE2);
	double det = rvE1.Dot(rvP);

	RVector rvT;
	if (det >0)
	{
		rvT = rvOrg - rv0;
	}
	else
	{
		rvT = rv0 - rvOrg;
		det = -det;
	}

	if (det < 0.0001)
		return false;

	if (!bPlaneCheckOnly)
	{
		*u = rvT.Dot(rvP);
		if (*u < 0.0 || *u > det)
			return false;

		RVector rvQ = rvT.Cross(rvE1);
		*v = rvDir.Dot(rvQ);

		if (*v < 0.0 || *u + *v > det)
			return false;

		*t = rvE2.Dot(rvQ);

		double inv_det = 1.0 / det;
		*t *= inv_det;
		*u *= inv_det;
		*v *= inv_det;
	}
	else
	{
		*u = rvT.Dot(rvP);

		RVector rvQ = rvT.Cross(rvE1);
		*v = rvDir.Dot(rvQ);

		*t = rvE2.Dot(rvQ);

		double inv_det = 1.0 / det;
		*t *= inv_det;
		*u *= inv_det;
		*v *= inv_det;
	}

	return true;
}

bool IntersectTriangle(RVector rvStart, RVector rvEnd, RVector rv0, RVector rv1, RVector rv2, RVector& rvInt)
{
	double t, u, v;
	bool b = IntersectTriangle(rvStart, rvEnd - rvStart, rv0, rv1, rv2, false, &t, &u, &v);
	if (b)
	{
		rvInt = rv0 + (rv1 - rv0)*u + (rv2 - rv0)*v;
	}
	return b;
}

bool IntersectPlane(RVector rvStart, RVector rvEnd, RVector rv0, RVector rv1, RVector rv2, RVector& rvInt)
{
	double t, u, v;
	bool b = IntersectTriangle(rvStart, rvEnd - rvStart, rv0, rv1, rv2, true, &t, &u, &v);
	if (b)
	{
		rvInt = rv0 + (rv1 - rv0)*u + (rv2 - rv0)*v;
		return true;
	}
	return false;
}

//ж2DµεǷ˳ʱ
bool Is2DTriangleClockwise(RVector rv0, RVector rv1, RVector rv2)
{
	double d1 = rv0.getAngleTo(rv1);
	double d2 = rv0.getAngleTo(rv2);

	if (fabs(d1 - d2) > M_PI)
		return d1 > d2;
	else
		return d1 < d2;
}

//ȡ2Dε
double Get2DTriangleArea(RVector& rv0, RVector& rv1, RVector& rv2)
{
	double a = rv0.getDistanceTo(rv1);
	double b = rv1.getDistanceTo(rv2);
	double c = rv2.getDistanceTo(rv0);
	double p = (a + b + c) / 2;
	double da = sqrt(p * (p - a)*(p - b)*(p - c));
	return da;
}

//ȡ2Dıε
double Get2DQuadArea(RVector& rvLt, RVector& rvRt, RVector& rvLb, RVector& rvRb)
{
	return Get2DTriangleArea(rvLt, rvRt, rvLb) + Get2DTriangleArea(rvRt, rvRb, rvLb);
}

//ȡ2Dıε
double Get2DQuadArea(RVector& rvLt, RVector& rvRt, RVector& rvLb, RVector& rvRb, RVector& rvM)
{
	return Get2DTriangleArea(rvLt, rvRt, rvM) + Get2DTriangleArea(rvRt, rvRb, rvM) + Get2DTriangleArea(rvLb, rvRb, rvM) + Get2DTriangleArea(rvLt, rvLb, rvM);
}

bool IsLineInteractionXaxis(double x1, double x2, double y, RVector& rv0, RVector& rv1)
{
	if (y<ov_min(rv0.y, rv1.y) || y>ov_max(rv0.y, rv1.y))
		return false;
	RVector rv = rv1;
	rv.scale((y - rv0.y) / (rv1.y - rv0.y), rv0);
	if (rv.x >= x1 && rv.x <= x2)
		return true;
	return false;
}

bool IsLineInteractionYaxis(double x, double y1, double y2, RVector& rv0, RVector& rv1)
{
	if (x<ov_min(rv0.x, rv1.x) || x>ov_max(rv0.x, rv1.x))
		return false;
	RVector rv = rv1;
	rv.scale((x - rv0.x) / (rv1.x - rv0.x), rv0);
	if (rv.y >= y1 && rv.y <= y2)
		return true;
	return false;
}

bool IsPtIn2DQuad(RVector& rvLt, RVector& rvRt, RVector& rvLb, RVector& rvRb, double x, double y)
{
	double dminx = ov_min(rvLt.x, ov_min(rvRt.x, ov_min(rvLb.x, rvRb.x)));
	double dmaxx = ov_max(rvLt.x, ov_max(rvRt.x, ov_max(rvLb.x, rvRb.x)));
	double dminy = ov_min(rvLt.y, ov_min(rvRt.y, ov_min(rvLb.y, rvRb.y)));
	double dmaxy = ov_max(rvLt.y, ov_max(rvRt.y, ov_max(rvLb.y, rvRb.y)));

	if (x<dminx || x>dmaxx || y<dminy || y>dmaxy)
		return false;

	RVector rv(x, y);
	RVector rvRight(dmaxx + 1, y);

	RVector rvInt;
	int nInt = 0;
	if (IsLineInteractionXaxis(x, dmaxx + 1, y, rvLt, rvRt))
		nInt++;
	if (IsLineInteractionXaxis(x, dmaxx + 1, y, rvRt, rvRb))
		nInt++;
	if (IsLineInteractionXaxis(x, dmaxx + 1, y, rvRb, rvLb))
		nInt++;
	if (IsLineInteractionXaxis(x, dmaxx + 1, y, rvLb, rvLt))
		nInt++;

	if (nInt % 2 == 0)
		return false;
	return true;
}

bool Is2DQuadIntersectBox(RVector& rvLt, RVector& rvRt, RVector& rvLb, RVector& rvRb, double dxLt, double dyLt, double dxRb, double dyRb)
{
	if (rvLt.x >= dxLt && rvLt.x <= dxRb && rvLt.y >= dyLt && rvLt.y <= dyRb)
		return true;
	if (rvRt.x >= dxLt && rvRt.x <= dxRb && rvRt.y >= dyLt && rvRt.y <= dyRb)
		return true;
	if (rvLb.x >= dxLt && rvLb.x <= dxRb && rvLb.y >= dyLt && rvLb.y <= dyRb)
		return true;
	if (rvRb.x >= dxLt && rvRb.x <= dxRb && rvRb.y >= dyLt && rvRb.y <= dyRb)
		return true;

	double dminx = ov_min(rvLt.x, ov_min(rvRt.x, ov_min(rvLb.x, rvRb.x)));
	double dmaxx = ov_max(rvLt.x, ov_max(rvRt.x, ov_max(rvLb.x, rvRb.x)));
	double dminy = ov_min(rvLt.y, ov_min(rvRt.y, ov_min(rvLb.y, rvRb.y)));
	double dmaxy = ov_max(rvLt.y, ov_max(rvRt.y, ov_max(rvLb.y, rvRb.y)));

	bool bIntersection = false;
	if (dxLt >= dminx && dxLt <= dmaxx && dyLt >= dminy && dyLt <= dmaxy)
		bIntersection = true;
	if (dxLt >= dminx && dxLt <= dmaxx && dyRb >= dminy && dyRb <= dmaxy)
		bIntersection = true;
	if (dxRb >= dminx && dxRb <= dmaxx && dyLt >= dminy && dyLt <= dmaxy)
		bIntersection = true;
	if (dxRb >= dminx && dxRb <= dmaxx && dyRb >= dminy && dyRb <= dmaxy)
		bIntersection = true;
	if (!bIntersection)
		return false;

	//жboxĽûrvLt, rvRt, rvLb, rvRb
	if (IsPtIn2DQuad(rvLt, rvRt, rvLb, rvRb, dxLt, dyLt))
		return true;
	if (IsPtIn2DQuad(rvLt, rvRt, rvLb, rvRb, dxRb, dyLt))
		return true;
	if (IsPtIn2DQuad(rvLt, rvRt, rvLb, rvRb, dxLt, dyRb))
		return true;
	if (IsPtIn2DQuad(rvLt, rvRt, rvLb, rvRb, dxRb, dyRb))
		return true;

	//жrvLt-rvRtboxıûཻ
	if (IsLineInteractionXaxis(dxLt, dxRb, dyLt, rvLt, rvRt))
		return true;
	if (IsLineInteractionXaxis(dxLt, dxRb, dyRb, rvLt, rvRt))
		return true;
	if (IsLineInteractionYaxis(dxLt, dyLt, dyRb, rvLt, rvRt))
		return true;
	if (IsLineInteractionYaxis(dxRb, dyLt, dyRb, rvLt, rvRt))
		return true;
	//жrvRt-rvRbboxıûཻ
	if (IsLineInteractionXaxis(dxLt, dxRb, dyLt, rvRt, rvRb))
		return true;
	if (IsLineInteractionXaxis(dxLt, dxRb, dyRb, rvRt, rvRb))
		return true;
	if (IsLineInteractionYaxis(dxLt, dyLt, dyRb, rvRt, rvRb))
		return true;
	if (IsLineInteractionYaxis(dxRb, dyLt, dyRb, rvRt, rvRb))
		return true;
	//жrvLb-rvRbboxıûཻ
	if (IsLineInteractionXaxis(dxLt, dxRb, dyLt, rvLb, rvRb))
		return true;
	if (IsLineInteractionXaxis(dxLt, dxRb, dyRb, rvLb, rvRb))
		return true;
	if (IsLineInteractionYaxis(dxLt, dyLt, dyRb, rvLb, rvRb))
		return true;
	if (IsLineInteractionYaxis(dxRb, dyLt, dyRb, rvLb, rvRb))
		return true;
	//жrvLt-rvLbboxıûཻ
	if (IsLineInteractionXaxis(dxLt, dxRb, dyLt, rvLt, rvLb))
		return true;
	if (IsLineInteractionXaxis(dxLt, dxRb, dyRb, rvLt, rvLb))
		return true;
	if (IsLineInteractionYaxis(dxLt, dyLt, dyRb, rvLt, rvLb))
		return true;
	if (IsLineInteractionYaxis(dxRb, dyLt, dyRb, rvLt, rvLb))
		return true;

	return false;
}

//ԲġXYɵBOXȡBoxеһеı
RVector GetPointScaleRvInBox(RVector rvCenter, RVector rvX, RVector rvY, RVector rvPoint)
{
	RVector rv;

	double dx = rvX.getDistanceTo(rvCenter);
	double dy = rvY.getDistanceTo(rvCenter);

	double dd = rvPoint.getDistanceTo(rvCenter);
	double dAngle = rvPoint.getAngleTo(rvCenter) - rvX.getAngleTo(rvCenter);

	double ddy = dd*sin(dAngle);
	double ddx = dd*cos(dAngle);

	rv.x = ddx / dx;
	rv.y = ddy / dy;

	if (rv.x < -1)
		rv.x = -1;
	if (rv.x > 1)
		rv.x = 1;
	if (rv.y < -1)
		rv.y = -1;
	if (rv.y > 1)
		rv.y = 1;

	return rv;
}

RVector GetPointInBoxByScaleRv(RVector rvCenter, RVector rvX, RVector rvY, RVector rvScale)
{
	RVector rv;

	double dx = rvX.getDistanceTo(rvCenter);
	double dy = rvY.getDistanceTo(rvCenter);

	rv.x = dx*rvScale.x + rvCenter.x;
	rv.y = dy*rvScale.y + rvCenter.y;

	double dAngle = rvX.getAngleTo(rvCenter);
	rv.rotate(dAngle + M_PI, rvCenter);

	return rv;
}



bool GetVerticalPedal(double x1, double y1, double x2, double y2, double xx, double yy, double &px, double &py)
{
	// 10:41 2017/1/16
	// ̫ᵼ, ƫ
	const double xxBak = xx;
	const double yyBak = yy;

	x1 -= xxBak;
	y1 -= yyBak;
	x2 -= xxBak;
	y2 -= yyBak;
	xx = yy = 0;

	// ȡߵĴ, ע: ԭֱߵĴ
	double cx = (x1 + x2) / 2;
	double cy = (y1 + y2) / 2;
	// double kk = atan2(y2 - y1, x2 - x1);

	// ʽ

	double a = 0, b = 0, c = 0;
	// б
	double kk = 0;
	if (fabs(y2 - y1) < 1e-6)
	{
		// бʲ
		a = -1;
		c = cx;
	}
	else
	{
		kk = -(x2 - x1) / (y2 - y1);

		// y-y0 = k(x - x0)
		a = kk;
		b = -1;
		c = cy - kk*cx;
	}

	px = (b*b*xx - a*b*yy - a*c) / (a*a + b*b);
	py = (a*a*yy - a*b*xx - b*c) / (a*a + b*b);

	px += xxBak;
	py += yyBak;

	return true;
}

bool GetCircelCenterByPoint3(double x1, double y1, double x2, double y2, double x3, double y3, double &cx, double &cy)
{
	// 㲻ܹ

	// 10:41 2017/1/16
	// ̫ᵼ, ƫ
	const double x1Bak = x1;
	const double y1Bak = y1;
	x1 = y1 = 0;
	x2 -= x1Bak;
	y2 -= y1Bak;
	x3 -= x1Bak;
	y3 -= y1Bak;

	double a = 2 * (x2 - x1);
	double b = 2 * (y2 - y1);
	double c = x2*x2 + y2*y2 - x1*x1 - y1*y1;
	double d = 2 * (x3 - x2);
	double e = 2 * (y3 - y2);
	double f = x3*x3 + y3*y3 - x2*x2 - y2*y2;

	double dd = b*d - e*a;
	if (fabs(dd) < 1e-6)
		return false;

	cx = (b*f - e*c) / dd;
	cy = (d*c - a*f) / dd;

	cx += x1Bak;
	cy += y1Bak;

	return true;
}

//	: ȡԹָΧڵֵ
double GetCycleRangeNum(double fNum, double fMin, double fMax, double dCycle)
{
	if (fMax <= fMin)
		return fNum;

	if (dCycle <= 1e-6)
	{
		dCycle = fMax - fMin;
	}
	else if (fabs(fMax - fMin - dCycle) > 1e-6)
	{
		// ΧպΪһ, Ȼû
		return fNum;
	}

	// ԸպõڱȽСǸֵ, ܵڱȽϴǸֵ
	while (fNum >= fMax)
	{
		fNum -= dCycle;
	}
	while (fNum < fMin)
	{
		fNum += dCycle;
	}

	return fNum;
}

double GetDegDiff(double d1, double d2)
{
	// ǶȵĲ
	// ʼΪ󲻳180

	double dd = fabs(d1 - d2);
	while (dd > 180)
	{
		if (d1 > d2)
			d1 -= 360;
		else
			d1 += 360;
		dd = fabs(d1 - d2);
	}

	return dd;
}

void PtAddRvBox(RVector& rvLt, RVector& rvRb, double x, double y)
{
	if (rvLt.x > x)
		rvLt.x = x;
	if (rvRb.x < x)
		rvRb.x = x;

	if (rvLt.y > y)
		rvLt.y = y;
	if (rvRb.y < y)
		rvRb.y = y;
}
